package owg.lgm.gl;

import com.jogamp.common.util.IntIntHashMap;
import com.jogamp.common.util.IntObjectHashMap;
import com.jogamp.opengl.GL2ES1;

import owg.lgm.IImage;
import owg.lgm.gl.BMFont.Char;
import owg.lgm.gl.BMFont.KerningPair;
import owg.lgm.util.AutoDestructibleImpl;

public class BMFontRenderer extends AutoDestructibleImpl
{
	public final String name;
	public final BMFont font;
	public final Tex2D[] pages;
	public final IntObjectHashMap chars;
	public final IntObjectHashMap kern;
	
	public BMFontRenderer(GL2ES1 gl, String name, BMFont font, IImage[] pageImages)
	{
		this.name = name;
		this.font = font;
		
		pages = new Tex2D[font.common.pages];
		for(int i = 0; i<pages.length; i++)
		{
			pages[i] = new Tex2D(gl, pageImages[i].toAlpha(), false, TexSampler.NEAREST_CLAMP);
			autoDestroy(pages[i]);
		}
		
		chars = new IntObjectHashMap(1+font.chars.length);
		for(int i = 0; i<font.chars.length; i++)
		{
			Char c = font.chars[i];
			chars.put(c.id, c);
		}
		
		if(font.kerningPairs != null)
		{
			kern = new IntObjectHashMap(1+font.kerningPairs.length);
			for(int i = 0; i<font.kerningPairs.length; i++)
			{
				KerningPair k = font.kerningPairs[i];
				IntIntHashMap k2 = new IntIntHashMap(32);
				k2.setKeyNotFoundValue(0);
				k2.put(k.second, k.amount);
				kern.put(k.first, k2);
			}
		}
		else
			kern = new IntObjectHashMap(1);
	}
	
	public int getWidth(CharSequence str, int maxWidth)
	{
		if(maxWidth <= 0)
			maxWidth = Integer.MAX_VALUE;
		int lineEnd = findNextLine(str, 0, maxWidth);
		int width = 0;
		int cx = 0;
		int rw = 0;
		char last = 0;
		for(int i = 0; i<str.length(); i++)
		{
			char ch = str.charAt(i);
			if(i == lineEnd)
			{
				lineEnd = findNextLine(str, i, maxWidth);
				width = Math.max(rw, width);
				cx = 0;
				rw = 0;
				last = 0;
			}
			Char c = (Char) chars.get(ch);
			if(c != null)
			{
				IntIntHashMap k1 = (IntIntHashMap) kern.get(last);
				if(k1 != null)
					cx += k1.get(ch);
				rw = cx + c.width;
				cx += c.xadvance;
			}
			last = ch;
		}
		return Math.max(rw, width);
	}
	
	protected int lineWidthImpl(CharSequence str, int lineStart, int lineEnd)
	{
		int cx = 0;
		int rw = 0;
		char last = 0;
		for(int i = lineStart; i<lineEnd; i++)
		{
			char ch = str.charAt(i);
			Char c = (Char) chars.get(ch);
			if(c != null)
			{
				IntIntHashMap k1 = (IntIntHashMap) kern.get(last);
				if(k1 != null)
					cx += k1.get(ch);
				rw = cx + c.width;
				cx += c.xadvance;
			}
			last = ch;
		}
		return rw;
	}
	
	protected int findNextLine(CharSequence str, int lineStart, int maxWidth)
	{
		int cx = 0;
		char last = 0;
		
		int chkI = 0;
		
		for(int i = lineStart; i<str.length(); i++)
		{
			char ch = str.charAt(i);
			if(ch == '\n')
				return i+1;
			Char c = (Char) chars.get(ch);
			if(c != null)
			{
				IntIntHashMap k1 = (IntIntHashMap) kern.get(last);
				if(k1 != null)
					cx += k1.get(ch);
				if(Character.isWhitespace(ch))
					chkI = i;
				else if(cx + c.width > maxWidth)
				{
					if(chkI > 0)
						return chkI+1;
					else
						return i;
				}
				cx += c.xadvance;
			}
			last = ch;
		}
		return str.length();
	}
	
	public int getHeight(CharSequence str, int lineHeight, int maxWidth)
	{
		if(lineHeight < 0)
			lineHeight = font.common.lineHeight;
		if(maxWidth <= 0)
			maxWidth = Integer.MAX_VALUE;
		return getHeightImpl(str, lineHeight, maxWidth);
	}
	
	protected int getHeightImpl(CharSequence str, int lineHeight, int maxWidth)
	{	
		int lineEnd = findNextLine(str, 0, maxWidth);
		int lines = 1;
		while(lineEnd < str.length())
		{
			lines++;
			lineEnd = findNextLine(str, lineEnd, maxWidth);
		}
		return lines * lineHeight;
	}


	public TextImpl build(GL2ES1 gl, CharSequence str, int lineHeight, int maxWidth, float halign, float valign)
	{
		TextImpl r = new TextImpl(gl, this, str.length(), BufferUsage.STATIC_DRAW);
		build(gl, r, str, lineHeight, maxWidth, halign, valign);
		return r;
	}
	public void build(GL2ES1 gl, Text output, CharSequence str, int lineHeight, int maxWidth, float halign, float valign)
	{
		if(maxWidth <= 0)
			maxWidth = Integer.MAX_VALUE;
		if(lineHeight < 0)
			lineHeight = font.common.lineHeight;
		
		output.prepare(str.length());
		
		int lineEnd = findNextLine(str, 0, maxWidth);
		int lineWidth = lineWidthImpl(str, 0, lineEnd);
		int stringHeight = getHeightImpl(str, lineHeight, maxWidth);
		int cx = -(int)(lineWidth*halign);
		int cy = -(int)(stringHeight*valign);
		char last = 0;
		for(int i = 0; i<str.length(); i++)
		{
			char ch = str.charAt(i);
			if(i == lineEnd)
			{
				lineEnd = findNextLine(str, lineEnd, maxWidth);
				lineWidth = lineWidthImpl(str, i, lineEnd);
				cx = -(int)(lineWidth*halign);
				cy += lineHeight;
				last = 0;
			}
			Char c = (Char) chars.get(ch);
			if(c != null)
			{
				IntIntHashMap k1 = (IntIntHashMap) kern.get(last);
				if(k1 != null)
					cx += k1.get(ch);
				
				int x0 = cx + c.xoffset;
				int y0 = cy + c.yoffset;
				int x1 = x0 + c.width;
				int y1 = y0 + c.height;
				
				int cx1 = c.x+c.width;
				int cy1 = c.y+c.height;
				
				output.putChar(str, i, c.page, x0, y0, (float)c.x/font.common.scaleW, (float)c.y/font.common.scaleH, x1, y1, (float)cx1/font.common.scaleW, (float)cy1/font.common.scaleH);
				
				cx += c.xadvance;
				last = ch;
			}
		}
		
		output.finish();
	}
}
