package owg.lgm.loader;

import owg.lgm.util.AutoDestructibleImpl;
import owg.lgm.util.FILE;
import owg.lgm.util.Letterbox;

import java.io.IOException;
import java.util.Stack;

import com.jogamp.opengl.GL2ES1;

import owg.lgm.Activity;
import owg.lgm.Background;
import owg.lgm.Fonts;
import owg.lgm.GMLib;
import owg.lgm.IImage;
import owg.lgm.ResourceException;
import owg.lgm.Sprite;
import owg.lgm.gl.BlendMode;
import owg.lgm.gl.Tex2D;
import owg.lgm.gl.TexSampler;
import owg.lgm.plaf.PNGIO;
import owg.lgm.plaf.Sound;
import owg.lgm.room.ActivityChangeType;
import owg.lgm.room.RoomData;
import owg.util.math.Lerp;

import static owg.lgm.plaf.Engine.EN;

public class LoadingScreen extends AutoDestructibleImpl implements Activity
{
	protected static class ProgFile
	{
		double p0, p1;
		FILE f;
		public ProgFile(double p0, double p1, FILE f)
		{
			this.p0 = p0;
			this.p1 = p1;
			this.f = f;
		}
	}
	
	IImage splashImg, fillerImg, progressImg;
	Tex2D splashTex, fillerTex, progressTex;
	float[] viewport = new float[4];
	
	double progress = 0.0;

	final ProgFile sprites, backgrounds, sounds, fonts, rooms;
	final ProgFile[] roots;
	final Stack<ProgFile> fileStack = new Stack<>();
	ProgFile currentRoot;
	Runnable endAction;
	
	Thread roomLoader;

	public LoadingScreen(Runnable endAction) throws IOException
	{
		if(EN.gameConfig.splashImage != null)
			splashImg = PNGIO.load(GMLib.file("res/"+EN.gameConfig.splashImage, false));
		if(EN.gameConfig.fillerImage != null)
			fillerImg = PNGIO.load(GMLib.file("res/"+EN.gameConfig.fillerImage, false));
		if(EN.gameConfig.progressImage != null)
			progressImg = PNGIO.load(GMLib.file("res/"+EN.gameConfig.progressImage, false));

		double p = 0.0;
		rooms = new ProgFile(p, p=0.05, GMLib.file("res/Rooms", false));
		fonts = new ProgFile(p, p=0.1, GMLib.file("res/Fonts", false));
		sounds = new ProgFile(p, p=0.2, GMLib.file("res/Sounds", false));
		backgrounds = new ProgFile(p, p=0.3, GMLib.file("res/Backgrounds", false));
		sprites = new ProgFile(p, p=1.0, GMLib.file("res/Sprites", false));

		roots = new ProgFile[] {sprites, backgrounds, sounds, fonts, rooms};

		for(int i = 0; i<roots.length; i++)
			fileStack.push(roots[i]);
		
		roomLoader = new Thread( () -> RoomData.readDir(rooms.f) );
		roomLoader.setDaemon(true);
		
		this.endAction = endAction;
	}

	public int getWidth()
	{
		if(splashImg != null)
			return splashImg.width;
		else
			return 256;
	}

	public int getHeight()
	{
		if(splashImg != null)
			return splashImg.height;
		else
			return 64;
	}
	
	@Override
	public void start(GL2ES1 gl, ActivityChangeType eventType)
	{
		if(splashImg != null)
		{
			splashTex = new Tex2D(gl, splashImg, false, TexSampler.NEAREST_REPEAT);
			autoDestroy(splashTex);
		}
		if(fillerImg != null) 
		{
			fillerTex = new Tex2D(gl, fillerImg, false, TexSampler.NEAREST_REPEAT);
			autoDestroy(fillerTex);
		}
		if(progressImg != null)
		{
			progressTex = new Tex2D(gl, progressImg, false, TexSampler.NEAREST_REPEAT);
			autoDestroy(progressTex);
		}
	}

	@Override
	public void run(GL2ES1 gl)
	{
		long t0 = System.currentTimeMillis();
		while(System.currentTimeMillis() < t0+16 && !fileStack.isEmpty())
		{
			ProgFile current = fileStack.pop();
			try {
				if(current == rooms)
					roomLoader.start();
				else
				{
					for(int i = 0; i<roots.length; i++)
					{
						if(current == roots[i])
							currentRoot = current;
					}
					try {
						FILE[] fs = current.f.listFiles();
						for(int i = 0; i<fs.length; i++)
							fileStack.push(new ProgFile(
									Lerp.lerp(current.p0, current.p1, (double)(fs.length-1-i)/fs.length),
									Lerp.lerp(current.p0, current.p1, (double)(fs.length-i)/fs.length), 
									fs[i]));
					} catch (IOException err) {
						if(currentRoot == sprites) {
							Sprite.load(EN.gl, current.f);
						} else if(currentRoot == backgrounds)
							Background.load(gl, current.f);
						else if(currentRoot == fonts)
							Fonts.load(gl, current.f);
						else if(currentRoot == sounds)
							Sound.load(current.f, EN.gameConfig.polyphony);
					}
				}
				progress = current.p0;
			} catch(IOException e) {
				throw new ResourceException("Failure at "+current.f+":", e);
			}
		}

    	EN.clear();
    	
		int w = EN.getSurfaceWidth();
		int h = EN.getSurfaceHeight();
		gl.glViewport(0, 0, w, h);
		gl.glMatrixMode(GL2ES1.GL_PROJECTION);
		gl.glLoadIdentity();
		gl.glOrthof(0, w, h, 0, -1, 1);

		gl.glMatrixMode(GL2ES1.GL_MODELVIEW);
		gl.glLoadIdentity();
		
		gl.glEnable(GL2ES1.GL_BLEND);
		BlendMode.NORMAL.set();

		gl.glColor4f(1, 1, 1, 1);
		Letterbox.getViewport(viewport, 0, EN.getSurfaceWidth(), EN.getSurfaceHeight(), splashImg.width, splashImg.height);
		
		if(splashTex != null)
		{
			splashTex.begin();
			float iw = viewport[2];
			float ih = viewport[3];
			float x0 = viewport[0]; 
			float y0 = viewport[1];
			gl.glPushMatrix();
			gl.glTranslatef(x0, y0, 0);
			gl.glScalef(iw, ih, 1);
			gl.glMatrixMode(GL2ES1.GL_TEXTURE);
			gl.glScalef(splashTex.mx, splashTex.my, 1.0f);
			EN.nwRect.render(gl);
			gl.glLoadIdentity();
			gl.glMatrixMode(GL2ES1.GL_MODELVIEW);
			gl.glPopMatrix();
			splashTex.end();
		}
		
		if(fillerTex != null)
		{
			fillerTex.begin();
			float iw = (float)(viewport[2]*EN.gameConfig.progressW);
			float ih = (float)(viewport[3]*EN.gameConfig.progressH);
			float x0 = (float)(viewport[0]+viewport[2]*EN.gameConfig.progressX); 
			float y0 = (float)(viewport[1]+viewport[3]*EN.gameConfig.progressY);
			gl.glPushMatrix();
			gl.glTranslatef(x0, y0, 0);
			gl.glScalef(iw, ih, 1);
			gl.glMatrixMode(GL2ES1.GL_TEXTURE);
			gl.glScalef(fillerTex.mx, fillerTex.my, 1.0f);
			EN.nwRect.render(gl);
			gl.glLoadIdentity();
			gl.glMatrixMode(GL2ES1.GL_MODELVIEW);
			gl.glPopMatrix();
			fillerTex.end();
		}
		
		if(progressTex != null)
		{
			progressTex.begin();
			float iw = (float)(progress * viewport[2]*EN.gameConfig.progressW);
			float ih = (float)(viewport[3]*EN.gameConfig.progressH);
			float x0 = (float)(viewport[0]+viewport[2]*EN.gameConfig.progressX); 
			float y0 = (float)(viewport[1]+viewport[3]*EN.gameConfig.progressY);
			gl.glPushMatrix();
			gl.glTranslatef(x0, y0, 0);
			gl.glScalef(iw, ih, 1);
			gl.glMatrixMode(GL2ES1.GL_TEXTURE);
			if(EN.gameConfig.scaleProgressBar)
				gl.glScalef(progressTex.mx, progressTex.my, 1.0f);
			else
				gl.glScalef((float) (progressTex.mx*progress), progressTex.my, 1.0f);
			EN.nwRect.render(gl);
			gl.glLoadIdentity();
			gl.glMatrixMode(GL2ES1.GL_MODELVIEW);
			gl.glPopMatrix();
			progressTex.end();
		}
		
		if(fileStack.isEmpty() && !roomLoader.isAlive())
			endAction.run();
	}
	
	@Override
	public void end(GL2ES1 gl, ActivityChangeType eventType)
	{
		dispose();
	}
}
