package owg.lgm.plaf;

import java.io.PrintStream;
import java.util.Arrays;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.Queue;

import com.jogamp.opengl.GLAnimatorControl;
import com.jogamp.opengl.GLAutoDrawable;
import com.jogamp.opengl.GLException;

public class UltimateAnimator implements GLAnimatorControl, Runnable
{
	protected static final int STOPPED = 0;
	protected static final int ANIMATING = 1;
	protected static final int PAUSED = 2;

	protected int targetFPS;

	protected long fpsStartTime = Long.MIN_VALUE;
	protected long lastFPSUpdateTime = fpsStartTime;
	protected int lastFPSPeriod = 0;
	
	protected long frameTime = Long.MIN_VALUE;
	
	protected int totalFPSFrames = 0;
	protected int updateFPSFrames;
	protected int framesUntilFPSUpdate;
	protected PrintStream fpsOut = null;

	protected Object stateLock = new Object();
	protected int state = STOPPED;
	protected Thread thread = null;
	
	protected final boolean daemon;
	
	protected Queue<Runnable> runnables = new LinkedList<>();

	LinkedHashSet<GLAutoDrawable> drawables = new LinkedHashSet<>();
	UncaughtExceptionHandler exceptionHandler = null;

	/**Creates a new {@link UltimateAnimator} which will attempt to achieve the given number of frames per second for the given drawable(s).
	 * The FPS update interval is set to the target FPS.
	 * The animator thread may be a daemon thread, meaning it will not keep the JVM alive if there are no other non-daemon threads.
	 * Note {@link #start()} must be called in order to start the animator.*/
	public UltimateAnimator(int targetFPS, boolean daemon, GLAutoDrawable... drawables)
	{
		this.targetFPS = targetFPS;
		this.updateFPSFrames = targetFPS==0?60:targetFPS;
		this.framesUntilFPSUpdate = targetFPS;
		this.daemon = daemon;
		this.drawables.addAll(Arrays.asList(drawables));
		for(GLAutoDrawable d : drawables)
			d.setAnimator(this);//Needed or NEWT will make deadlocks
	}

	@Override
	public void setUpdateFPSFrames(int frames, PrintStream out)
	{
		synchronized (stateLock)
		{
			this.updateFPSFrames = frames;
			this.framesUntilFPSUpdate = frames;
			this.fpsOut = out;
			resetFPSCounter();
		}
	}

	@Override
	public void resetFPSCounter()
	{
		synchronized (stateLock)
		{
			fpsStartTime = System.currentTimeMillis();
			lastFPSUpdateTime = fpsStartTime;
			totalFPSFrames = 0;
			framesUntilFPSUpdate = updateFPSFrames;
			resetFrameTime();
		}
	}
	
	/**Reset the frame time to the current time, effectively erasing any accumulated timing errors. 
	 * It is useful to call this after known timing disruptions, such as after going to a new level, to prevent the animator from attempting to compensate.<br>
	 * <br>
	 * Note the timestamp is also automatically reset on a successful {@link #start()} or {@link #resume()}, and on {@link #resetFPSCounter()}.*/
	public void resetFrameTime()
	{
		synchronized (stateLock)
		{
			frameTime = System.nanoTime();	
		}
	}

	@Override
	public int getUpdateFPSFrames()
	{
		synchronized (stateLock)
		{
			return updateFPSFrames;
		}
	}

	@Override
	public long getFPSStartTime()
	{
		synchronized (stateLock)
		{
			return fpsStartTime;
		}
	}

	@Override
	public long getLastFPSUpdateTime()
	{
		synchronized (stateLock)
		{
			return lastFPSUpdateTime;
		}
	}

	@Override
	public long getLastFPSPeriod()
	{
		synchronized (stateLock)
		{
			return lastFPSPeriod;
		}
	}

	@Override
	public float getLastFPS()
	{
		synchronized (stateLock)
		{
			return lastFPSPeriod==0 ? 0 : (1000.0f * (float)updateFPSFrames/lastFPSPeriod);
		}
	}

	@Override
	public int getTotalFPSFrames()
	{
		synchronized (stateLock)
		{
			return totalFPSFrames;
		}
	}

	@Override
	public long getTotalFPSDuration()
	{
		synchronized (stateLock)
		{
			return System.currentTimeMillis() - fpsStartTime;
		}
	}

	@Override
	public float getTotalFPS()
	{
		synchronized (stateLock)
		{
			return (float)(totalFPSFrames)/(lastFPSUpdateTime-fpsStartTime);
		}
	}

	@Override
	public boolean isStarted()
	{
		synchronized (stateLock)
		{
			return state != STOPPED;
		}
	}

	@Override
	public boolean isAnimating()
	{
		synchronized (stateLock)
		{
			return state == ANIMATING;
		}
	}

	@Override
	public boolean isPaused()
	{
		synchronized (stateLock)
		{
			return state == PAUSED;
		}
	}

	@Override
	public Thread getThread()
	{
		synchronized (stateLock)
		{
			return thread;
		}
	}

	@Override
	public boolean start()
	{
		synchronized (stateLock)
		{
			if(state != STOPPED)
				return false;
			state = ANIMATING;
			thread = new Thread(this);
			resetFPSCounter();
			thread.setDaemon(daemon);
			thread.start();
			return true;
		}
	}

	@Override
	public boolean stop()
	{
		synchronized (stateLock)
		{
			if(state == STOPPED)
				return false;
			state = STOPPED;
			thread = null;
			return true;
		}
	}

	@Override
	public boolean pause()
	{
		synchronized (stateLock)
		{
			if(state != ANIMATING)
				return false;
			state = PAUSED;
			return true;
		}
	}

	@Override
	public boolean resume()
	{
		synchronized (stateLock)
		{
			if(state != PAUSED)
				return false;
			state = ANIMATING;
			resetFrameTime();
			return true;
		}
	}

	@Override
	public void add(GLAutoDrawable drawable)
	{
		synchronized (stateLock)
		{
			drawables.add(drawable);
			drawable.setAnimator(this);
		}
	}

	@Override
	public void remove(GLAutoDrawable drawable)
	{
		synchronized (stateLock)
		{
			drawables.remove(drawable);
			if(drawable.getAnimator() == this)
				drawable.setAnimator(null);
		}
	}

	public GLAutoDrawable[] getDrawables()
	{
		synchronized (stateLock)
		{
			return drawables.toArray(new GLAutoDrawable[drawables.size()]);
		}
	}

	@Override
	public UncaughtExceptionHandler getUncaughtExceptionHandler()
	{
		synchronized (stateLock)
		{
			return exceptionHandler;
		}
	}

	@Override
	public void setUncaughtExceptionHandler(UncaughtExceptionHandler handler)
	{
		synchronized (stateLock)
		{
			this.exceptionHandler = handler;
		}
	}

	@Override
	public void run()
	{
		while(frame()) {}
	}
	
	protected boolean frame()
	{
		UncaughtExceptionHandler exceptionHandler;
		GLAutoDrawable[] drawables;
		long waitTime;
		synchronized (stateLock)
		{
			if(state == STOPPED)
				return false;
			
			exceptionHandler = this.exceptionHandler;
			if(state == PAUSED)
			{
				drawables = null;
				//Limit amount of wasted CPU when paused
				if(targetFPS <= 0)
					waitTime = 16;
				else
					waitTime = 1000/targetFPS;
			}
			else
			{
				waitTime = 0;
				drawables = this.drawables.toArray(new GLAutoDrawable[this.drawables.size()]);
			}
		}

		if(drawables != null)
		{
			//The display does work in another thread, so holding the state lock here constitutes a deadlock
			for(GLAutoDrawable d : drawables)
			{
				try {
					if(d.isRealized())
						d.display();
					//Else window is likely closed
				} catch(Throwable e) {
					//Strip exception wrappers created by jogamp
					if(e instanceof RuntimeException && e.getCause() instanceof GLException &&
							e.getCause().getMessage().startsWith("Caught ") &&
							e.getCause().getCause() != null)
						e = e.getCause().getCause();
	
					if(e instanceof GLException && e.getCause() != null && 
							e.getMessage().startsWith("Caught "))
						e = e.getCause();
	
					if(exceptionHandler != null)
						exceptionHandler.uncaughtException(this, d, e);
					else if(e instanceof RuntimeException)
						throw (RuntimeException)e;
					else
						throw new RuntimeException(e);
					stop();
					return false;
				}
			}
			
			Runnable[] toRun;
			synchronized(runnables) {
				toRun = runnables.toArray(new Runnable[runnables.size()]);
				runnables.clear();
			}
			try {
				for(int i = 0; i<toRun.length; i++)
					toRun[i].run();
			} catch(Throwable e) {
				if(exceptionHandler != null)
					exceptionHandler.uncaughtException(this, null, e);
				else if(e instanceof RuntimeException)
					throw (RuntimeException)e;
				else
					throw new RuntimeException(e);
			}
			
			//FPS counter, implemented to GLAnimatorControl's spec
			synchronized (stateLock) {
				if(updateFPSFrames > 0)
				{
					framesUntilFPSUpdate--;
					if(framesUntilFPSUpdate <= 0)
					{
						long currentTime = System.currentTimeMillis();
						lastFPSPeriod = (int) (currentTime - lastFPSUpdateTime);
						lastFPSUpdateTime = currentTime;
						framesUntilFPSUpdate = updateFPSFrames;
						if(fpsOut != null)
							fpsOut.println(getLastFPS());
					}
				}
				
				//Error sensitive sleeps using nanosecond timestamps
				long nanoTime = System.nanoTime();
				//Discard error if it is too large
				if(Math.abs(frameTime-nanoTime) > 500_000_000L)
					frameTime = nanoTime;
				//The following division is exact enough
				//(millions of frames between significant rounding errors)
				if(targetFPS > 0)
					frameTime += 1_000_000_000L/targetFPS;
				waitTime = (frameTime - nanoTime)/1000000L;
			}
		}
		
		if(waitTime > 0)
		{
			try {
				Thread.sleep(waitTime);
			} catch (InterruptedException e) {
				//Don't care
			}
		}
		return true;
	}
	/**Invokes the given task on the animator thread after rendering.*/
	public void invokeOnAnimator(Runnable task)
	{
		synchronized (runnables) 
		{
			runnables.offer(task);
		}
	}
	/**Returns a {@link System#nanoTime()}-compatible timestamp indicating when the current frame was scheduled to be rendered.
	 * This value is far more stable than measuring wall time. Under normal circumstances, it will be an idealized, evenly spaced sequence.
	 * It will be disrupted by sustained drops in FPS due to performance problems.*/
	public long getFrameTime()
	{
		synchronized (stateLock) {
			return frameTime;
		}
	}
	
	public int getTargetFPS()
	{
		synchronized (stateLock) {
			return targetFPS;
		}
	}
	
	public void setTargetFPS(int fps)
	{
		synchronized (stateLock) {
			targetFPS = fps;
		}
	}
}
