package owg.competent;

import java.nio.Buffer;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.nio.FloatBuffer;
import java.nio.IntBuffer;
import java.nio.LongBuffer;
import java.nio.ShortBuffer;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import static owg.competent.CAErrorCode.*;

/**<p>
 * This class is a thin wrapper on top of {@link CANatives}, which adds documentation, exception handling, type safe enums, 
 * and reduces the need for direct buffers, while still accurately representing the low level CA API.
 *</p><p>
 *Competent Audio must be initialized before use (<code>caInitialize()</code>) and uninitialized after use (<code>caUninitialize()</code>).
 *Thread-safe and thread-unsafe modes are available.
 *</p><p>
 *CA uses a machine tree model. Audio data flows from the leaf machines towards the root machine.
 *Machines that produce audio data are called source machines. Machines that consume audio data are called destination machines.
 *</p><p>
 *Three machine types are included:<ol>
 *<li>Sinks, which are machine tree roots. They are destination machines that typically feed audio data to a set of speakers.</li>
 *<li>Samplers, which are machine tree leaves. They are source machines that produce audio data from sample data.</li>
 *<li>Mixers, which can be installed between other machines in the tree. A mixer is a source and destination machine.
 *   They combine multiple audio data streams into one, while offering some options for manipulating the audio.</li>
 *</ol>
 *</p><p>
 *Machines are represented as <code>long</code> handles that are actually pointers.
 *As such, improper handling of this API may cause memory corruption and segmentation faults.
 *</p><p>
 *{@link CompetentAudio} is an object. This design allows applications to substitute a custom implementation for purposes like debugging.
 *This is not an invitation to create multiple instances, which would normally be pointless.
 *</p><p>
 *A singleton instance is available at {@link CompetentAudio#CA}, which can be static imported for convenience.
 *</p>*/
public class CompetentAudio
{
	public final List<AudioBackend> backends = Collections.unmodifiableList(Arrays.asList(AudioBackend.values()));
	public final List<SamplerState> samplerStates = Collections.unmodifiableList(Arrays.asList(SamplerState.values()));
	public final List<DSPErrorCode> dspErrorCodes = Collections.unmodifiableList(Arrays.asList(DSPErrorCode.values()));
	public final List<DSPRoutine> dspRoutines = Collections.unmodifiableList(Arrays.asList(DSPRoutine.values()));
	
	public static CompetentAudio CA = new CompetentAudio();
	
	protected static Object initLock = new Object();
	protected static boolean threadSafe = true;
	protected static Appendable debugStream = null;
	protected static boolean initialized = false;
	protected static CADeletionListenerList uninitializeListeners = new CADeletionListenerList(initLock);
	
	/**<p>
	 * Can be called while CA is not initialized to change which dispatcher is used: </ul>
	 *<li> true - Thread-safe dispatcher (application may access CA from multiple threads)</li>
	 *<li> false - Thread-unsafe dispatcher (application must access CA from a single thread only)</li>
	 *</ul>
	 *The default is <code>true</code>.
	 *</p><p>
	 *Possible errors:<ul>
	 *<li>ALREADY_INITIALIZED if CA is currently initialized</li>
	 *</ul></p>*/
	public static void setThreadSafe(boolean threadSafe) throws CAException
	{
		synchronized (initLock) {
			if(initialized)
				throw new CAException(ALREADY_INITIALIZED);
			CompetentAudio.threadSafe = threadSafe;
		}
	}
	/**Returns true if thread safe mode is enabled (default: <code>true</code>).*/
	public static boolean isThreadSafe() {
		synchronized (initLock) {
			return threadSafe;
		}
	}
	/**<p>
	 * Can be called while CA is not initialized to change the debug stream.
	 * The native library will print debug information to this stream when unusual conditions occur.<br>
	 * <br>
	 * The default is <code>null</code>.
	 * </p><p>
	 * Possible errors:<ul>
	 * <li>ALREADY_INITIALIZED if CA is currently initialized</li>
	 * </ul></p>*/
	public static void setDebugStream(Appendable debugStream) throws CAException
	{
		synchronized (initLock) {
			if(initialized)
				throw new CAException(ALREADY_INITIALIZED);
			CompetentAudio.debugStream = debugStream;
		}
	}
	/**Returns the current debug stream (default: <code>null</code>).*/
	public static Appendable getDebugStream() {
		synchronized (initLock) {
			return debugStream;
		}
	}
	/**Returns true if CA is currently initialized.*/
	public static boolean isInitialized() {
		synchronized (initLock) {
			return initialized;
		}
	}
	
	/**Used to add or remove callbacks for when {@link #uninitialize()} is called.*/
	public static CADeletionListenerList uninitializeListeners() {
		return uninitializeListeners;
	}
	
	protected CompetentAudio() {}

	///////////////////////////////////////////
	/*Core functionality for Competent Audio.*/
	///////////////////////////////////////////

	/**Returns the version number of the currently loaded Competent Audio library code.*/
	public int getVersion()
	{
		return CANatives.getVersion();
	}

	/**<p>
	 *Must be called by the application before the rest of the functionality in this file can be used.
	 *Attempting to call any other function before this is called will result in undefined behaviour.
	 *</p><p>
	 *Unlike the native version, calling this function multiple times is not considered an error. 
	 *If CA is already initialized, <code>false</code> is returned.
	 *</p><p>
	 *The application may change the dispatcher to use by calling {@link #setThreadSafe(boolean)} before this function.
	 *</p><p>
	 *The application may assume that this function succeeds even if the system does not have functioning sound system.
	 *In such a case, only the Dummy device will be available, which behaves like a normal device except it does not 
	 *actually output sound to any speakers.
	 *</p><p>
	 *For diagnosing problems encountered during initialization, see the caGetBackendError function.
	 *</p><p>
	 *Possible errors:<ul>
	 *<li>INSUFFICIENT_MEMORY if there was not enough memory to initialize Competent Audio.</li>
	 *</ul></p>*/
	public boolean initialize() throws CAException
	{
		synchronized (initLock) {
			if(initialized)
				return false;
			CAException.throwIfError(CANatives.initialize(threadSafe, debugStream));
			initialized = true;
			return true;
		}
	}

	/**<p>
	 *Must be called by the application after it is done using the functionality in this file.
	 *The application should do its best to ensure this function is called before the program terminates.
	 *Calling this function will close all streams and free all memory currently in use by machine trees 
	 *(with the possible exception of sample data that is not freed by a dequeue callback). 
	 *All machine pointers are invalid after this call.
	 *</p><p>
	 *Unlike the native version, calling this function multiple times is not considered an error. 
	 *If CA is not currently initialized, <code>false</code> is returned.
	 *</p><p>
	 *Possible errors:<ul>
	 *<li>NOT_INITIALIZED if CA is not currently initialized</li>
	 *</ul></p>*/
	public boolean uninitialize() throws CAException
	{
		synchronized (initLock) {
			if(!initialized)
				return false;
			//Note: It is important to mark as deleted first, or CASink will produce a concurrent modification error
			initialized = false;
			uninitializeListeners.deleted(this);
			CAException.throwIfError(CANatives.uninitialize());
			return true;
		}
	}

	/**<p>Returns the error code generated while trying to connect to the given backend.
	 * </p><p>
	 *Possible return values:<ul>
	 *<li>NO_ERROR 						The connection to the given backend was successful.</li>
	 *<li>BACKEND_NOT_INITIALIZED 		No attempt was made to connect to the given backend,
	 *									because another backend was successfully initialized first.</li>
	 *<li>ERROR_BACKEND_UNAVAILABLE 	the libsoundio build does not contain support for the given backend,
	 *									typically because the backend is for another operating system.</li>
	 *<li>ERROR_MISSING_SHARED_OBJECT	A required shared object file (so or dll) failed to load.
										Most likely, this is because the backend is not installed on the system.</li>
	 *<li>ERROR_MISSING_SYMBOLS 		A shared object file (so or dll) was successfully loaded, 
										but was missing symbols (functions) that are required to use it. 
	 									This can happen if an incompatible version of the backend is installed.</li>
	 *<li>ERROR_INIT_AUDIO_BACKEND 		The backend is not operational, e.g. it is installed, but not enabled.</li>
	 *<li>ERROR_SYSTEM_RESOURCES 		Failed due to running out of some system resource other than memory,
										such as having too many connections already open.</li>
	 *<li>ERROR_BACKEND_DISCONNECTED	If the backend became disabled while trying to connect.</li>
	 *<li>ERROR_NO_SUCH_CLIENT			If the JACK connection was deleted while connecting.</li>
	 *</ul>
	 *Possible errors:<ul> 
	 *<li>INDEX_OUT_OF_BOUNDS			The given backend is an invalid index.</li>
	 *<li>NOT_INITIALIZED 				Competent Audio has not been initialized.</li>
	 *</ul></p>*/
	public CAErrorCode getBackendError(int backend) throws CAException
	{
		CAErrorCode err = getError(CANatives.getBackendError(backend));
		if(		err == NO_ERROR || err == BACKEND_NOT_INITIALIZED || 
				err == ERROR_BACKEND_UNAVAILABLE || err == ERROR_MISSING_SHARED_OBJECT || err == ERROR_MISSING_SYMBOLS || 
				err == ERROR_INIT_AUDIO_BACKEND || err == ERROR_SYSTEM_RESOURCES || 
				err == ERROR_BACKEND_DISCONNECTED || err == ERROR_NO_SUCH_CLIENT)
			return err;
		throw new CAException(err);
	}

	/**Can be passed as the device index to caOpenSink to automatically select the default output device.
	 *Sinks that use this device index will automatically switch to the new default output device if it changes.*/
	public static final int DEVICE_INDEX_AUTO = CANatives.DEVICE_INDEX_AUTO;
	
	/**<p>
	 *Returns the number of available audio devices.
	 *</p><p>
	 *Possible errors:<ul>
	 *<li>NOT_INITIALIZED if Competent Audio has not been initialized</li>
	 *</ul></p>*/
	public int getDeviceCount() throws CAException
	{
		return CAException.throwIfError(CANatives.getDeviceCount());
	}
	
	/**<p>
	 *Returns the index of the current default audio output device.
	 *If the system does not have a default device, then this returns the dummy device index.
	 *</p><p>
	 *Most applications should use DEVICE_INDEX_AUTO instead.
	 *</p><p>
	 *Possible errors:<ul>
	 *<li>NOT_INITIALIZED if Competent Audio has not been initialized</li>
	 *</ul></p>*/
	public int getDefaultDevice() throws CAException
	{
		return CAException.throwIfError(CANatives.getDefaultDevice());
	}

	/**<p>
	 *Returns the name of the audio device with the given index.
	 *The device index must be in the half-open range [0, getDeviceCount()&gt;, or alternatively DEVICE_INDEX_AUTO.
	 *The name of DEVICE_INDEX_AUTO will be reported as "Automatic".
	 *</p><p>
	 *Possible errors:<ul>
	 *<li>NOT_INITIALIZED Competent Audio has not been initialized</li>
	 *<li>INDEX_OUT_OF_BOUNDS The device index is out of bounds</li>
	 * </ul></p>*/
	public String getDeviceName(int deviceIndex) throws CAException
	{
		String name = CANatives.getDeviceName(deviceIndex);
		if(name == null)
		{
			int deviceCount = getDeviceCount();
			if(deviceIndex < DEVICE_INDEX_AUTO || deviceIndex >= deviceCount)
				throw new CAException(INDEX_OUT_OF_BOUNDS, 
						"deviceIndex="+deviceIndex+", deviceCount="+deviceCount);
			else
				throw new CAException(NOT_INITIALIZED);
		}
		return name;
	}
	
	/**<p>
	 *Returns the id of the audio device with the given index. This is suitable for storage in configuration files.
	 *The device index must be in the half-open range [0, getDeviceCount()&gt;, or alternatively DEVICE_INDEX_AUTO.
	 *The id of DEVICE_INDEX_AUTO will be reported as "auto".
	 *</p><p>
	 *Possible errors:<ul>
	 *<li>NOT_INITIALIZED Competent Audio is not initialized</li>
	 *<li>INDEX_OUT_OF_BOUNDS The device index is out of bounds</li>
	 * </ul></p>*/
	public String getDeviceID(int deviceIndex) throws CAException
	{
		String id = CANatives.getDeviceID(deviceIndex);
		if(id == null)
		{
			int deviceCount = getDeviceCount();
			if(deviceIndex < DEVICE_INDEX_AUTO || deviceIndex >= deviceCount)
				throw new CAException(INDEX_OUT_OF_BOUNDS, 
						"deviceIndex="+deviceIndex+", deviceCount="+deviceCount);
			else
				throw new CAException(NOT_INITIALIZED);
		}
		return id;
	}
	
	/**<p>
	 *Returns the backend type of the audio device with the given index.
	 *The device index must be in the half-open range [0, getDeviceCount()&gt;, or alternatively DEVICE_INDEX_AUTO.
	 *</p><p>
	 *Possible errors:<ul>
	 *<li>NOT_INITIALIZED Competent Audio is not initialized</li>
	 *<li>INDEX_OUT_OF_BOUNDS The device index is out of bounds</li>
	 * </ul></p>*/
	public AudioBackend getDeviceBackend(int deviceIndex) throws CAException
	{
		int backend = CAException.throwIfError(CANatives.getDeviceBackend(deviceIndex));
		if(backend < 0 || backend >= backends.size())
			throw new CAException(INCONSISTENT_DATA, "backend="+backend);
		return backends.get(backend);
	}

	/**<p>
	 *Refreshes the device list.
	 *Applications that allow the user to select an output device from a list may want to call this before displaying the list,
	 *and possibly provide a refresh button in the user interface.
	 *</p><p>
	 *Note: This function does not reconnect to the backend if the backend connection has failed (during or after initialization).
	 *</p><p>
	 *Possible errors:<ul> 
	 *<li>NOT_INITIALIZED if Competent Audio has not been initialized</li>
	 *</ul></p>*/
	public void refreshDeviceList() throws CAException
	{
		CAException.throwIfError(CANatives.refreshDeviceList());
	}

	/**<p>
	 *Deletes the given machine and all remaining machines that feed into it.
	 *On success, this returns NO_ERROR and frees the memory of the machine and all remaining inputs to it.
	 *Any remaining pointers to those machines will be invalid after this call.
	 *Note the memory is not necessarily freed immediately due to the threading model.
	 *</p><p>
	 *Possible errors:<ul>
	 *<li>NULL_POINTER if the machine pointer is NULL.</li>
	 *</ul></p>*/
	public void deleteMachine(long machine) throws CAException
	{
		CAException.throwIfError(CANatives.deleteMachine(machine));
	}

	/**<p>
	 *Creates a new unopened sink. caOpenSink() must be called to begin processing audio.
	 *</p><p>
	 *A sink must have a sound model. See {@link SoundModel#MONO} or {@link SoundModel#STEREO}.
	 *</p><p>
	 *maxChunkFrames specifies the maximum number of frames that may be processed as a chunk. The value must be positive. 
	 *This may affect both timing resolution and latency. A reasonable default is (requestedSampleRate/60).
	 *</p><p>
	 *Note that DSP implementations generally need to pre-allocate enough memory for at least one chunk of maximum size,
	 *i.e. frameSize * maxChunkFrames * 4 bytes, where frameSize is specified by the sound model. 
	 *ca_dsp has a memory limit of 65535 bytes; unusually high values may make ca_dsp unusable (see caSetDSP).
	 *</p><p>
	 *Higher values may cause higher latency as well as a decrease in the frequency of the message processing on the audio callback,
	 *leading to worse timing resolution. For instance, if the chunk size is 1/30th of the sample rate, then the time between the 
	 *processing of sample playback requests may be as high as 1/30th of a second.
	 *</p><p>
	 *Lower values increase the frequency of function calls on the audio callback, increasing CPU usage and potentially introducing audio gaps.
	 *</p><p>
	 *The actual chunk size depends on the underlying system, and may change (often) during playback. maxChunkFrames only enforces an upper limit.
	 *</p><p>
	 *Possible errors:<ul>
	 *<li>CA_NOT_INITIALIZED        if Competent Audio has not been initialized.</li>
	 *<li>CA_NULL_POINTER           if sinkPtr or soundModel is NULL.</li>
	 *<li>CA_INVALID_CHUNK_SIZE     if maxChunkSize is 0.</li>
	 *<li>CA_INSUFFICIENT_MEMORY    if there was not enough memory to create the sink.</li>
	 *</ul></p>*/
	public long newSink(SoundModel soundModel, int maxChunkFrames) throws CAException
	{
		LongBuffer sinkPtr = ByteBuffer.allocateDirect(8).asLongBuffer();
		CAException.throwIfError(CANatives.newSink(sinkPtr, soundModel.id, maxChunkFrames));
		return sinkPtr.get();
	}

	/**<p>
	 *Opens the given sink with the given requested parameters. If the sink is already open, then it is closed first.
	 *</p><p>
	 *The device index must be in the half-open range [0, caGetDeviceCount()>, or alternatively CA_DEVICE_INDEX_AUTO. 
	 *Most applications should use CA_DEVICE_INDEX_AUTO.
	 *Note that opening multiple sinks for the same device will not work for all devices. 
	 *</p><p>
	 *Most applications should request a sample rate of 44100. 
	 *There is no guarantee that the requested sample rate will be used, but it does not matter for most applications,
	 *since CA resamples clips automatically. If you need to know, call caGetSampleRate.
	 *</p><p>
	 *The actual sample format and speaker layout are decided by the system. This does not matter for most applications,
	 *since CA performs format and speaker layout conversions automatically. 
	 *If you need to know the speaker count, call caGetSpeakerCount.
	 *</p><p>
	 *The device refuses to open, then a dummy stream is opened instead.
	 *A dummy stream ensures the application logic will not experience major disruptions, but will not output sound to any actual device.
	 *</p><p>
	 *If using CA_DEVICE_INDEX_AUTO, and the default device changes, then the sink is automatically reopened.
	 *If the stream parameters change as a result of this, then DSPs are re-initialized (via CADestinationFunc.streamParametersChanged).
	 *The application will normally not need to take any special precautions for this.
	 *</p><p>
	 *This function always returns CA_NO_ERROR unless it is used incorrectly, runs out of memory, or suffers an unexpected failure.
	 *However, if a stream for the requested device cannot be opened, then a dummy stream is opened instead, and the sink error state is set.
	 *The sink error state can be queried by calling caGetSinkError().
	 *</p><p>
	 *Possible errors:<ul>
	 *<li>CA_NULL_POINTER           if the machine pointer is NULL.</li>
	 *<li>CA_WRONG_MACHINE_TYPE     if the given machine is not a sink.</li>
	 *<li>CA_INDEX_OUT_OF_BOUNDS    if the device index is out of bounds.</li>
	 *<li>CA_INSUFFICIENT_MEMORY    if there was not enough memory to open the sink.</li>
	 *</ul></p>*/
	public void openSink(long sinkMachine, int requestedDeviceIndex, int requestedSampleRate) throws CAException
	{
		CAException.throwIfError(CANatives.openSink(sinkMachine, requestedDeviceIndex, requestedSampleRate));
	}

	/**<p>
	 *If the given sink (or the sink associated with the given machine) is in an error state, 
	 *then this returns an error code describing the state. This can happen if: 
	 * - caOpenSink fails to open the requested device
	 * - the requested device is disconnected or disabled during playback
	 * - the default device switches to an incompatible device when using DEVICE_INDEX_AUTO
	 *When in an error state, the sink will normally use a dummy stream, which processes audio like a normal stream,
	 *but does not output the audio to any actual device.
	 *</p><p>
	 *Possible return values:<ul>
	 *<li>NO_ERROR 						the given sink is not in an error state.</li>
	 *<li>ERROR_STREAMING 				the stream was forcibly closed after it had started.</li>
	 *<li>ERROR_SAMPLE_FORMAT 			the device does not support any of the formats defined by CASampleFormat.</li>
	 *<li>ERROR_INCOMPATIBLE_DEVICE 	the device driver reported support for a format it does not actually support.</li>
	 *<li>ERROR_NO_SUCH_DEVICE 			the audio device was disconnected or disabled before it could be opened.</li>
	 *<li>ERROR_BACKEND_DISCONNECTED	the system's audio service was shut down.</li>
	 *<li>ERROR_NO_SUCH_CLIENT 			the application was removed from the client list in the JACK audio server.</li>
	 *<li>ERROR_OPENING_DEVICE 			the device could not be opened for another reason, such as too many streams being open.</li>
	 *</ul>
	 *</p><p>
	 *Possible errors:<ul>
	 *<li>NULL_POINTER 					the machine pointer is NULL.</li>
	 *<li>WRONG_MACHINE_TYPE 			the given machine is not a sink.</li>
	 *</ul></p>*/
	public CAErrorCode getSinkError(long sinkMachine) throws CAException
	{
		CAErrorCode err = getError(CANatives.getSinkError(sinkMachine));
		if(err == NO_ERROR || err == ERROR_STREAMING || err == ERROR_SAMPLE_FORMAT || 
				err == ERROR_INCOMPATIBLE_DEVICE || err == ERROR_NO_SUCH_DEVICE || err == ERROR_BACKEND_DISCONNECTED || 
				err == ERROR_NO_SUCH_CLIENT || err == ERROR_OPENING_DEVICE)
			return err;
		throw new CAException(err);
	}

	/**<p>
	 *Gets the actual sample rate of the given sink (or that of the sink associated with the given machine).
	 *</p><p>
	 *Possible errors:<ul>
	 *<li>NULL_POINTER 			if the machine pointer is NULL.</li>
	 *<li>ERROR_SAMPLE_RATE 	if the sink is not open.</li>
	 *</ul></p>*/
	public int getSampleRate(long machine) throws CAException
	{
		return CAException.throwIfError(CANatives.getSampleRate(machine));
	}

	/**<p>
	 *Gets the actual speaker count of the given sink (or that of the sink associated with the given machine).
	 *</p><p>
	 *Possible errors:<ul>
	 *<li>NULL_POINTER 			if the machine pointer is NULL.</li>
	 *<li>ERROR_SPEAKER_LAYOUT 	if the sink is not open.</li>
	 *</ul></p>*/
	public int getSpeakerCount(long machine) throws CAException
	{
		return CAException.throwIfError(CANatives.getSpeakerCount(machine));
	}

	/**<p>
	 *Resets the pull function for the given destination machine if it is different from the default, and frees the old one.
	 *Typically, this is used to remove a signal processor attached to a mixer or a sink.
	 *</p><p>
	 *Possible errors:<ul>
	 *<li>NULL_POINTER 			if dst is NULL.</li>
	 *<li>NOT_A_DESTINATION 	if dst is not a destination machine.</li>
	 *<li>INSUFFICIENT_MEMORY	if there was not enough memory to allocate a new default function container</li>
	 *</ul></p>*/
	public void resetDestinationFunc(long dst) throws CAException
	{
		CAException.throwIfError(CANatives.resetDestinationFunc(dst));
	}

	/**<p>
	 *Sets the gain multiplier for the given machine.
	 *</p><p> 
	 *Gain multipliers are used to adjust the playback volume of a machine (and all of its inputs).
	 *A value of 1 indicates the normal playback volume. Any finite multiplier is valid.
	 *</p><p>
	 *The gain will fade to the new value over the number of seconds given by fadeTime.
	 *If fadeTime is nonpositive, the change will be instantaneous.
	 *</p><p>
	 *Possible errors:<ul>
	 *<li>CA_NULL_POINTER if machine is NULL.</li>
	 *<li>CA_INVALID_GAIN if the gain multiplier is an infinity or NaN.</li>
	 *<li>CA_INSUFFICIENT_MEMORY if there was not enough memory to perform the change.</li>
	 *</ul></p>*/
	public void setGain(long machine, double fadeTime, float gain) throws CAException
	{
		CAException.throwIfError(CANatives.setGain(machine, fadeTime, gain));
	}

	/**<p>
	 *Sets the frequency multiplier for the given machine.
	 *Frequency multipliers are used to adjust the playback speed of a machine (and all of its inputs).
	 *This has the effect of altering the perceived tempo and pitch of the audio.
	 *</p><p>
	 *The frequency will fade to the new value over the number of seconds given by fadeTime.
	 *If fadeTime is nonpositive, the change will be instantaneous.
	 *</p><p>
	 *Possible errors:<ul>
	 *<li>NULL_POINTER 			if machine is NULL.</li>
	 *<li>INVALID_FREQUENCY 	if the frequency multiplier is negative.</li>
	 *<li>INSUFFICIENT_MEMORY 	if there was not enough memory to perform the change.</li>
	 *</ul></p>*/
	public void setFrequency(long machine, double fadeTime, double frequency) throws CAException
	{
		CAException.throwIfError(CANatives.setFrequency(machine, fadeTime, frequency));
	}



	////////////////////////////////////////
	/*Competent Audio Mixer functionality.*/
	////////////////////////////////////////

	/**Mixers combine multiple audio signals to one.
	 *
	 *They can be used to logically group sounds, and modulate their gain and frequency at a single point.
	 *
	 *Furthermore, they can be used as arbitrary signal processing machines (by attaching a custom destination function).*/



	/**<p>
	 *Creates a new mixer connected to the given destination machine.
	 *</p><p>
	 *Possible errors:<ul>
	 *<li>NULL_POINTER 			if dst is NULL.</li>
	 *<li>NOT_A_DESTINATION 	if dst is not a destination machine.</li>
	 *<li>INSUFFICIENT_MEMORY 	if there is not enough memory to create the mixer.</li>
	 *</ul></p>*/
	public long newMixer(long dst) throws CAException
	{
		LongBuffer mixerPtr = ByteBuffer.allocateDirect(8).asLongBuffer();
		CAException.throwIfError(CANatives.newMixer(mixerPtr, dst));
		return mixerPtr.get();
	}



	//////////////////////////////////////////
	/*Competent Audio Sampler functionality.*/
	//////////////////////////////////////////

	/**A sampler produces audio data from an audio clip, or a queue thereof.
	 *When a sampler is stopped, it is deactivated and consumes no processing resources.
	 *
	 *Clips can be dynamically queued up for gapless playback or streaming.
	 *
	 *Any number of channels is supported. Applications can control how those channels are mixed using the channel matrix.*/



	/**<p>
	 *Creates a new sampler bound to the given destination machine, with no clips queued.
	 *At least one clip must be enqueued for the sampler to become useful.
	 *</p><p>
	 *Possible errors:<ul>
	 *<li>CA_NULL_POINTER if samplerPtr or dst is NULL.</li>
	 *<li>CA_NOT_A_DESTINATION if dst is not a destination machine.</li>
	 *<li>CA_INSUFFICIENT_MEMORY if there is not enough memory to create the sampler.</li>
	 *</ul></p>*/
	public long newSampler(long dst) throws CAException
	{
		LongBuffer samplerPtr = ByteBuffer.allocateDirect(8).asLongBuffer();
		CAException.throwIfError(CANatives.newSampler(samplerPtr, dst));
		return samplerPtr.get();
	}
	
	/**<p>
	 *Enables or disables interpolation for the given sampler.
	 *Interpolation is on by default.
	 *</p><p>
	 *Possible errors:<ul>
	 *<li>NULL_POINTER 				if sampler is NULL</li>
	 *<li>INCORRECT_MACHINE_TYPE 	if sampler is not a sampler machine.</li>
	 *</ul></p>*/
	public void setInterpolation(long sampler, boolean enable) throws CAException
	{
		CAException.throwIfError(CANatives.setInterpolation(sampler, enable));
	}

	/**<p>
	 *Switches the state of the given sampler to PLAYING.
	 *</p><p>
	 *Possible errors:<ul>
	 *<li>NULL_POINTER 				if sampler is NULL</li>
	 *<li>INCORRECT_MACHINE_TYPE 	if sampler is not a sampler machine.</li>
	 *</ul></p>*/
	public void play(long sampler) throws CAException
	{
		CAException.throwIfError(CANatives.play(sampler));
	}
	
	/**<p>
	 *Switches the state of the given sampler to CA_PLAYING, starting at the given position in seconds.
	 *This function calculates positions using only the clip data; frequency scale factors are ignored.
	 *</p><p>
	 *If the position is negative, it is clamped to zero.
	 *If the position is beyond the current length of the clip queue, then: <ul>
	 *<li>If the sampler is looping, it starts playing from zero.</li>
	 *<li>Otherwise, the sampler is stopped.</li>
	 *</ul></p><p>
	 *Possible errors:<ul>
	 *<li>NULL_POINTER 				if sampler is NULL</li>
	 *<li>INCORRECT_MACHINE_TYPE 	if sampler is not a sampler machine.</li>
	 *<li>CA_INSUFFICIENT_MEMORY 	if there is not enough memory to perform the change.</li>
	 *</ul></p>*/
	public void playAt(long sampler, double startPos) throws CAException
	{
		CAException.throwIfError(CANatives.playAt(sampler, startPos));
	}

	/**<p>
	 *Switches the state of the given sampler to STOPPED.
	 *</p><p>
	 *Possible errors:<ul>
	 *<li>NULL_POINTER 				if sampler is NULL</li>
	 *<li>INCORRECT_MACHINE_TYPE 	if sampler is not a sampler machine.</li>
	 *</ul></p>*/
	public void stop(long sampler) throws CAException
	{
		CAException.throwIfError(CANatives.stop(sampler));
	}

	/**<p>
	 *Switches the state of the given sampler to PAUSED if it is playing.
	 *If it is not playing, then this does nothing.
	 *</p><p>
	 *Possible errors:<ul>
	 *<li>NULL_POINTER 				if sampler is NULL</li>
	 *<li>INCORRECT_MACHINE_TYPE 	if sampler is not a sampler machine.</li>
	 *</ul></p>*/
	public void pause(long sampler) throws CAException
	{
		CAException.throwIfError(CANatives.pause(sampler));
	}

	/**<p>
	 *Switches the state of the given sampler to PLAYING if it is paused.
	 *If it is not paused, then this does nothing.
	 *Use this if you do not want the sound to restart if it is not actually paused.
	 *</p><p>
	 *Possible errors:<ul>
	 *<li>NULL_POINTER 				if sampler is NULL</li>
	 *<li>INCORRECT_MACHINE_TYPE 	if sampler is not a sampler machine.</li>
	 *</ul></p>*/
	public void resume(long sampler) throws CAException
	{
		CAException.throwIfError(CANatives.resume(sampler));
	}

	/**<p>
	 *Switches the state of the given sampler to CA_PLAYING.
	 *Note that unlike caPlay, this function does not rewind the sound if it is already playing.
	 *This is particularly useful for streaming samplers that might spuriously stop due to running out of clips.
	 *</p><p>
	 *Possible errors:<ul>
	 *<li>NULL_POINTER 				if sampler is NULL</li>
	 *<li>INCORRECT_MACHINE_TYPE 	if sampler is not a sampler machine.</li>
	 *</ul></p>*/
	public void ensurePlaying(long sampler) throws CAException
	{
		CAException.throwIfError(CANatives.ensurePlaying(sampler));
	}

	/**<p>
	 *Gets the current state for the given sampler.
	 *</p><p>
	 *Possible errors:<ul>
	 *<li>NULL_POINTER 				if sampler is NULL</li>
	 *<li>INCORRECT_MACHINE_TYPE 	if sampler is not a sampler machine.</li>
	 *</ul></p>*/
	public SamplerState getState(long sampler) throws CAException
	{
		int samplerState = CAException.throwIfError(CANatives.getState(sampler));
		if(samplerState < 0 || samplerState >= samplerStates.size())
			throw new CAException(INCONSISTENT_DATA, "samplerState="+samplerState);
		return samplerStates.get(samplerState);
		
	}
	
	/**<p>
	 *Enables or disables looping for the given sampler.
	 *Looping is off by default.
	 *</p><p>
	 *Possible errors:<ul>
	 *<li>NULL_POINTER 				if sampler is NULL</li>
	 *<li>INCORRECT_MACHINE_TYPE 	if sampler is not a sampler machine.</li>
	 *</ul></p>*/
	public void setLooping(long sampler, boolean looping) throws CAException
	{
		CAException.throwIfError(CANatives.setLooping(sampler, looping));
	}
	
	/**<p>
	 *Returns whether the given sampler has looping enabled.
	 *</p><p>
	 *Possible errors:<ul>
	 *<li>NULL_POINTER 				if sampler is NULL</li>
	 *<li>INCORRECT_MACHINE_TYPE 	if sampler is not a sampler machine.</li>
	 *</ul></p>*/
	public boolean getLooping(long sampler) throws CAException
	{
		return CAException.throwIfError(CANatives.getLooping(sampler)) != 0;
	}
	
	/**<p>
	 *Sets the loop restart position in seconds for the given sampler.
	 *If the given position is negative, it is clamped to zero.
	 *If loop start position is out of bounds, it behaves as if it is zero.
	 *The default loop start position is zero.
	 *</p><p>
	 *Possible errors:<ul>
	 *<li>NULL_POINTER 				if sampler is NULL</li>
	 *<li>INCORRECT_MACHINE_TYPE 	if sampler is not a sampler machine.</li>
	 *<li>CA_INSUFFICIENT_MEMORY 	if there was not enough memory to perform the change.</li>
	 *</ul></p>*/
	public void setLoopStart(long sampler, double loopStart) throws CAException
	{
		CAException.throwIfError(CANatives.setLoopStart(sampler, loopStart));
	}
	
	/**<p>
	 *Returns the current loop restart position in seconds (default zero).
	 *</p><p>
	 *Possible errors:<ul>
	 *<li>NULL_POINTER 				if sampler is NULL</li>
	 *<li>INCORRECT_MACHINE_TYPE 	if sampler is not a sampler machine.</li>
	 *</ul></p>*/
	public double getLoopStart(long sampler) throws CAException
	{
		double result = CANatives.getLoopStart(sampler);
		if(result < 0)
			throw new CAException(CAErrorCode.getError((int)result));
		return result;
	}

	/**<p>
	 *Enqueues a clip to the given sampler.
	 *</p><p>
	 *autoDequeue says whether the clip should be automatically removed from the sampler's clip queue when it has played to completion.
	 *copyData specifies whether the sample data should be copied by the sampler. 
	 *If not, then the clip's sampleData buffer must not be modified by the application after this call,
	 *and <b>its contents will be converted to native-endian if this is not already the case.</b>
	 *</p><p>
	 *Possible errors:<ul>
	 *<li>CA_NULL_POINTER           if the machine pointer is NULL, or if sampleData is NULL.</li>
	 *<li>CA_WRONG_MACHINE_TYPE     if the machine is not a sampler.</li>
	 *<li>INSUFFICIENT_MEMORY 		if there is not enough memory to enqueue the clip.</li>
	 *</ul></p>*/
	public void enqueueClip(long sampler, boolean autoDequeue, boolean copyData, CAClip clip) throws CAException
	{
		CAException.throwIfError(CANatives.enqueueClip(sampler, autoDequeue, copyData, 
				clip.getChannelCount(), clip.getSampleRate(), 
				clip.getSampleFormat().ordinal(), clip.byteOrder==ByteOrder.LITTLE_ENDIAN, 
				clip.getFrameCount(), clip.getSampleData()));
		if(!copyData)
			clip.setByteOrder(ByteOrder.nativeOrder());
	}

	/**<p>
	 *Dequeues all clips for the given sampler, and stops it if it is playing.
	 *You can use this if you want to reuse a sampler for a new set of clips.
	 *</p><p>
	 *Possible errors:<ul>
	 *<li>CA_NULL_POINTER if the machine pointer is null.</li>
	 *<li>CA_WRONG_MACHINE_POINTER if the machine is not a sampler.</li>
	 *</ul></p>*/
	public void clearClipQueue(long sampler) throws CAException
	{
		CAException.throwIfError(CANatives.clearClipQueue(sampler));
	}

	/**<p>
	 *Returns the number of clips currently in the queue for the given sampler.
	 *If you want to limit the number of clips in the queue (e.g. for memory or timing reasons), 
	 *then you can use this value to determine if you should enqueue another clip or wait.
	 *</p><p>
	 *Possible errors:<ul>
	 *<li>CA_NULL_POINTER if the machine pointer is null.</li>
	 *<li>CA_WRONG_MACHINE_POINTER if the machine is not a sampler.</li>
	 *</ul></p>*/
	public int getClipQueueSize(long sampler) throws CAException
	{
		return CAException.throwIfError(CANatives.getClipQueueSize(sampler));
	}

	/**<p>
	 *Returns the number of clips that have been played to completion for the given sampler.
	 *This property is reset when the sampler is stopped.
	 *You can use this function to obtain some information about the playback position of a multi-clip sampler.
	 *</p><p>
	 *Possible errors:<ul>
	 *<li>CA_NULL_POINTER if the machine pointer is null.</li>
	 *<li>CA_WRONG_MACHINE_POINTER if the machine is not a sampler.</li>
	 *</ul></p>*/
	public int getProcessedClips(long sampler) throws CAException
	{
		return CAException.throwIfError(CANatives.getProcessedClips(sampler));
	}

	/*A channel matrix provides control over how sample channels are mapped to output channels.
	 *The matrix consists of a series of channel vectors. Each channel vector consists of one or more 32 bit floats.
	 *The size and layout of a channel vector is given by the current sound model, 
	 *which is the sound model of the sink at the root of the sampler's machine tree.
	 *
	 *For instance, if you are using a stereo sound model, then channel vectors have 2 components: [gain, pan].
	 *
	 *Normally, you want to provide one channel vector for each sample channel, so if you are using stereo audio clips,
	 *then your matrix should have 2 channel vectors. In the stereo sound model, this means 2*2 = 4 elements.
	 *
	 *Channel vectors can be used to implement positional audio, or a simple dynamic music system, e.g. by using one
	 *sample channel for each instrument, and modulating them individually with the gain components.*/


	/**<p>
	 *Resets the channel mixing mode for the given sampler to the default.
	 *</p><p>
	 *Possible errors:<ul>
	 *<li>CA_NULL_POINTER if the machine pointer is null.</li>
 	 *<li>CA_WRONG_MACHINE_POINTER if the machine is not a sampler.</li>
	 *</ul></p>*/
	public void clearChannelMatrix(long sampler) throws CAException
	{
		CAException.throwIfError(CANatives.clearChannelMatrix(sampler));
	}

	/**
	 * <p>
	 *Sets the channel matrix for the given sampler.
	 *</p><p>
	 *Possible errors:<ul>
	 *<li>CA_NULL_POINTER if the machine pointer is null.</li>
	 *<li>CA_WRONG_MACHINE_POINTER if the machine is not a sampler.</li>
	 *<li>CA_BAD_CHANNEL_VECTOR_SIZE if the vector size does not match that of the current sound model.</li>
	 *<li>CA_INSUFFICIENT_MEMORY if there is not enough memory to allocate the mixing matrix.</li>
	 *</ul></p>
	 * @param sampler The sampler to apply the channel matrix to
	 * @param fadeTime Each element in the matrix will fade from the current to the new value over this number of seconds.
     *      If fadeTime is nonpositive, or if there was no channel matrix applied already, 
     *      then the change is instantaneous. Also, if the new matrix has fewer channels than the old matrix,
     *      then the missing channels will not fade (they are cut off instantly).
	 * @param channelCount The number of channel vectors in the matrix. 
	 * 		Typically this matches the number of channels in the audio clip(s) used by this sampler. 
	 * 		If the matrix is larger, then the extra vectors have no effect.
	 *      If the matrix is smaller, then the sample channels not covered will be silenced.
	 * @param vectorSize The size of each channel vector. 
	 * 		This must match the vector size specified by the current sound model.
	 * @param matrix The channel matrix, as an array of channel vectors. 
	 * 		See the sound model documentation for a definition of the format and semantics of channel vectors. 
	 * 		If this is NULL, then this function behaves like caClearChannelMatrix.
	 */
	public void setChannelMatrix(long sampler, double fadeTime, int channelCount, int vectorSize, FloatBuffer matrix) throws CAException
	{
		CAException.throwIfError(CANatives.setChannelMatrix(sampler, fadeTime, channelCount, vectorSize, matrix));
	}

	////////////////////////////////////////////////
	/*RISVM-based general purpose signal processor*/
	////////////////////////////////////////////////

	/**Returns the version number of the currently loaded ca_dsp library code.*/
	public int getDSPVersion() throws CAException
	{
		return CANatives.getDSPVersion();
	}

	/**<p>
	 *Attaches a digital signal processor (DSP) to the machine associated with the given destination.
	 *The program and message are copied by this function.
	 *</p><p>
	 *Possible errors:<ul>
	 *<li>CA_NULL_POINTER if the destination or program pointer was NULL, or if messageSize>0 and message is NULL</li>
	 *<li>CA_WRONG_DESTINATION_FUNC if there is no signal processor attached to the destination</li>
	 *<li>CA_DSP_TOO_LARGE if the DSP cannot fit within the 16 bit address space</li>
	 *<li>CA_INSUFFICIENT_MEMORY if there was not enough memory to allocate the DSP</li>
	 *</ul></p>*/
	public void setDSP(long dst, short programSize, ByteBuffer program, short stackSize, int messageSize, Buffer message) throws CAException
	{
		CAException.throwIfError(CANatives.setDSP(dst, programSize, program, stackSize, messageSize, message));
	}

	/**<p>
	 *Passes the given message to the DSP attached to the given machine. 
	 *If the DSP is in an error state, this does nothing.
	 *</p><p>
	 *Possible errors:<ul>
	 *<li>CA_NULL_POINTER if the destination was NULL, or if messageSize>0 and message is NULL</li>
	 *<li>CA_WRONG_DESTINATION_FUNC if there is no signal processor attached to the destination</li>
	 *<li>CA_INSUFFICIENT_MEMORY if there was not enough memory to allocate a copy of the message</li>
	 *</ul></p>*/

	public void sendDSPMessage(long dst, int userArgsSize, Buffer userArgs) throws CAException
	{
		CAException.throwIfError(CANatives.sendDSPMessage(dst, userArgsSize, userArgs));
	}

	/**<p>
	 * If an error occurs in a DSP program, then it is stopped and not invoked again. This function returns the
	 * error state of the DSP attached to the given machine, or null if the DSP is not in an error state.
	 *</p><p>
	 *Possible errors:<ul>
	 *<li>NULL_POINTER if the machine argument is NULL.</li>
	 *<li>NOT_A_DESTINATION if the given machine is not a destination machine.</li>
	 *<li>WRONG_DESTINATION_FUNC if no signal processor is attached to the given machine.</li>
	 *</ul></p>*/
	public DSPErrorState getDSPError(long machine) throws CAException
	{
		int dspError = CAException.throwIfError(CANatives.getDSPError(machine, null, null, null, null));
		if(dspError == 0)
			return null;
		if(dspError >= dspErrorCodes.size())
			throw new CAException(INCONSISTENT_DATA, "dspError="+dspError);
		
		//  1*uint64_t 	invocationIndex
		// 20*int32_t 	registers
		//  1*uint8_t 	traceSize
		//255*uint16_t 	trace
		ByteBuffer buffer = ByteBuffer.allocateDirect(1+20*4+1+255*2);
		ByteBuffer routinePtr = buffer.slice();
		buffer.position(1);
		IntBuffer registerPtr = buffer.slice().asIntBuffer();
		buffer.position(1+20*4);
		ByteBuffer traceSizePtr = buffer.slice();
		buffer.position(1+20*4+1);
		ShortBuffer tracePtr = buffer.slice().asShortBuffer();
		
		int r = CAException.throwIfError(CANatives.getDSPError(machine, routinePtr, registerPtr, traceSizePtr, tracePtr));
		if(r != dspError)
			throw new RuntimeException(r+"!="+dspError);
		short[] stackTrace = new short[traceSizePtr.get()&0xFF];
		tracePtr.get(stackTrace);
		int[] registers = new int[20];
		registerPtr.get(registers);
		byte dspRoutine = routinePtr.get();
		if(dspRoutine < 0 || dspRoutine >= dspRoutines.size())
			throw new CAException(INCONSISTENT_DATA, "dspRoutine="+dspRoutine);
		return new DSPErrorState(dspErrorCodes.get(dspError), dspRoutines.get(dspRoutine), registers, stackTrace);
	}
}
