package owg.competent.soundfile;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.util.ArrayList;

import owg.competent.CAClip;
import owg.competent.util.CAUtil;

/**Audio file format support.*/
public abstract class CAClipLoader 
{
	/**Declares support for a type of audio file.*/
	public interface AudioFileHandler
	{
		/**Returns true if this handler supports the given file.
		 * Typically this just checks the file extension.*/
		public abstract boolean supports(File f);
		/**Returns an {@link AudioDecoderStream} for the given file.*/
		public abstract AudioDecoderStream open(File f) throws IOException;
	}
	
	/**The current list of audio file handlers. 
	 * Add more items here to allow {@link #read(File)} to support more file types.
	 * The handlers are consulted in reverse order, so the last handler has the highest priority.*/
	public static final ArrayList<AudioFileHandler> AUDIO_FILE_HANDLERS;
	static
	{
		AUDIO_FILE_HANDLERS = new ArrayList<AudioFileHandler>();
		AUDIO_FILE_HANDLERS.add(new AudioFileHandler() {
			@Override
			public boolean supports(File f) {
				return f.getPath().endsWith(".wav") || f.getPath().endsWith(".WAV");
			}
			@Override
			public AudioDecoderStream open(File f) throws IOException {
				return new WavInputStream(new FileInputStream(f));
			}
		});
		AUDIO_FILE_HANDLERS.add(new AudioFileHandler() {
			@Override
			public boolean supports(File f) {
				return f.getPath().endsWith(".ogg") || f.getPath().endsWith(".OGG");
			}
			@Override
			public AudioDecoderStream open(File f) throws IOException {
				return new OggInputStream(new FileInputStream(f));
			}
		});
		AUDIO_FILE_HANDLERS.add(new AudioFileHandler() {
			@Override
			public boolean supports(File f) {
				return f.getPath().endsWith(".mp3") || f.getPath().endsWith(".MP3");
			}
			@Override
			public AudioDecoderStream open(File f) throws IOException {
				return new MP3InputStream(new FileInputStream(f));
			}
		});
	}
	
	/**{@link ByteArrayOutputStream} that exposes the backing array for performance considerations.*/
	public static class ExposedByteArrayOutputStream extends ByteArrayOutputStream {
		/**Creates an {@link ExposedByteArrayOutputStream} with the given initial size in bytes. It will grow on demand.*/
		public ExposedByteArrayOutputStream(int initialSize) {
			super(initialSize);
		}
		/**Returns the current backing array of this stream. <i>The returned array must not be modified while the stream is in use.
		  *The backing array may or may not be replaced with a new array when writing to the stream.</i>*/
		public byte[] getArray()
		{
			return buf;
		}
	}
	
	/**
	 * Tries to read the given file using one of the currently available {@link #AUDIO_FILE_HANDLERS}.
	 * <br><br>
	 * This method <b>returns <code>null</code></b> if the file was not recognized by any of the handlers.
	 * @param f The file to try to read
	 * @return The loaded clip, or null if the file was not recognized as a supported audio file.
	 * @throws IOException If the file was recognized by a handler, but there was an error reading it.
	 */
	public static CAClip read(File f) throws IOException
	{
		for(int i = AUDIO_FILE_HANDLERS.size()-1; i>=0; i--)
		{
			AudioFileHandler s = AUDIO_FILE_HANDLERS.get(i);
			if(s.supports(f))
			{
				AudioDecoderStream in = null;
				try {
					in = s.open(f);
					ByteBuffer data;
					byte[] buf = new byte[4096];
					int size = 0;
					
					if(in instanceof SizedAudioDecoderStream)
					{
						int expectedSize = ((SizedAudioDecoderStream)in).getSize();
						data = ByteBuffer.allocateDirect(expectedSize);
						while(size < expectedSize)
						{
							int read = in.read(buf);
							if(read == -1)
								throw new IOException("Unexpected end of file: "+f);
							data.put(buf, 0, read);
							size += read;
						}
					}
					else
					{
						ExposedByteArrayOutputStream bos = new ExposedByteArrayOutputStream(4096);
						while(true)
						{
							int read = in.read(buf);
							if(read == -1)
								break;
							bos.write(buf, 0, read);
							size += read;
						}
						data = ByteBuffer.allocateDirect(size);
						data.put(bos.getArray(), 0, size);
					}
					data.order(in.isLittleEndian() ? 
							ByteOrder.LITTLE_ENDIAN : ByteOrder.BIG_ENDIAN);
					return new CAClip(in.getChannelCount(), in.getSampleRate(), in.getSampleFormat(), data);
				} finally {
					CAUtil.close(in);
				}
			}
		}
		return null;
	}
	
	private CAClipLoader() {}
}
